#!/usr/bin/env python3 -B
# coding=utf-8

"""
Toshiba COM Extract
Toshiba BIOS COM Extractor
Copyright (C) 2018-2024 Plato Mavropoulos
"""

import os
import subprocess

from biosutilities.common.externals import comextract_path
from biosutilities.common.paths import is_file, make_dirs, path_stem, safe_name
from biosutilities.common.patterns import PAT_TOSHIBA_COM
from biosutilities.common.system import printer
from biosutilities.common.templates import BIOSUtility
from biosutilities.common.texts import file_to_bytes


class ToshibaComExtract(BIOSUtility):
    """ Toshiba BIOS COM Extractor """

    TITLE: str = 'Toshiba BIOS COM Extractor'

    def check_format(self, input_object: str | bytes | bytearray) -> bool:
        """ Check if input is Toshiba BIOS COM image """

        input_buffer: bytes = file_to_bytes(in_object=input_object)

        return bool(PAT_TOSHIBA_COM.search(string=input_buffer, endpos=0x100))

    def parse_format(self, input_object: str | bytes | bytearray, extract_path: str, padding: int = 0) -> bool:
        """ Parse & Extract Toshiba BIOS COM image """

        make_dirs(in_path=extract_path, delete=True)

        if isinstance(input_object, str) and is_file(in_path=input_object):
            input_path: str = input_object
        else:
            input_path = os.path.join(extract_path, 'toshiba_bios.com')

            with open(file=input_path, mode='wb') as input_buffer:
                input_buffer.write(file_to_bytes(in_object=input_object))

        output_name: str = f'{safe_name(in_name=path_stem(in_path=input_path))}_extracted.bin'

        output_path: str = os.path.join(extract_path, output_name)

        try:
            subprocess.run([comextract_path(), input_path, output_path], check=True, stdout=subprocess.DEVNULL)

            if not is_file(in_path=output_path):
                raise FileNotFoundError('EXTRACTED_FILE_MISSING')
        except Exception as error:  # pylint: disable=broad-except
            printer(message=f'Error: ToshibaComExtractor could not extract {input_path}: {error}!', padding=padding)

            return False

        if input_path != input_object:
            os.remove(path=input_path)

        printer(message='Successful extraction via ToshibaComExtractor!', padding=padding)

        return True


if __name__ == '__main__':
    ToshibaComExtract().run_utility()



#!/usr/bin/env python3 -B
# coding=utf-8

"""
Copyright (C) 2022-2024 Plato Mavropoulos
"""

import os
import subprocess

from typing import Final

from biosutilities.common.externals import szip_path, tiano_path
from biosutilities.common.paths import is_dir, is_empty_dir
from biosutilities.common.system import printer

# 7-Zip switches to auto rename, ignore passwords, ignore prompts, ignore wildcards,
# eliminate root duplication, set UTF-8 charset, suppress stdout, suppress stderr,
# suppress progress, disable headers, disable progress, disable output logs
SZIP_COMMON: Final[list[str]] = ['-aou', '-p', '-y', '-spd', '-spe', '-sccUTF-8',
                                 '-bso0', '-bse0', '-bsp0', '-ba', '-bd', '-bb0']

# Success exit codes (0 = OK, 1 = Warnings)
SZIP_SUCCESS: Final[list[int]] = [0, 1]


def szip_code_assert(exit_code: int) -> None:
    """ Check 7-Zip bad exit codes (0 OK, 1 Warning) """

    if exit_code not in SZIP_SUCCESS:
        raise ValueError(f'Bad exit code: {exit_code}')


def szip_switches(in_switches: list[str]) -> list[str]:
    """ Generate 7-Zip command line switches """

    common_switches: list[str] = SZIP_COMMON

    for in_switch in in_switches:
        for sw_pattern in ('-p', '-ao', '-bs', '-bb', '-scc'):
            if in_switch.startswith(sw_pattern):
                common_switches = [sw for sw in common_switches if not sw.startswith(sw_pattern)]

                break

    return [*set(common_switches + in_switches), '--']


def is_szip_supported(in_path: str, padding: int = 0, args: list | None = None, silent: bool = True) -> bool:
    """ Check if file is 7-Zip supported """

    try:
        if args is None:
            args = []

        szip_c: list[str] = [szip_path(), 't', *szip_switches(in_switches=[*args]), in_path]

        szip_t: subprocess.CompletedProcess[bytes] = subprocess.run(args=szip_c, check=False)

        szip_code_assert(exit_code=szip_t.returncode)
    except Exception as error:  # pylint: disable=broad-except
        if not silent:
            printer(message=f'Error: 7-Zip could not check support for file {in_path}: {error}!', padding=padding)

        return False

    return True


def szip_decompress(in_path: str, out_path: str, in_name: str = 'archive', padding: int = 0, args: list | None = None,
                    check: bool = False, silent: bool = False) -> bool:
    """ Archive decompression via 7-Zip """

    try:
        if args is None:
            args = []

        szip_c: list[str] = [szip_path(), 'x', *szip_switches(in_switches=[*args, f'-o{out_path}']), in_path]

        szip_x: subprocess.CompletedProcess[bytes] = subprocess.run(args=szip_c, check=False)

        if check:
            szip_code_assert(exit_code=szip_x.returncode)

        if not (is_dir(in_path=out_path) and not is_empty_dir(in_path=out_path)):
            raise OSError(f'Extraction directory is empty or missing: {out_path}')
    except Exception as error:  # pylint: disable=broad-except
        if not silent:
            printer(message=f'Error: 7-Zip could not extract {in_name} file {in_path}: {error}!', padding=padding)

        return False

    if not silent:
        printer(message=f'Successful {in_name} decompression via 7-Zip!', padding=padding)

    return True


def efi_compress_sizes(data: bytes | bytearray) -> tuple[int, int]:
    """ Get EFI compression sizes """

    size_compress: int = int.from_bytes(bytes=data[0x0:0x4], byteorder='little')

    size_original: int = int.from_bytes(bytes=data[0x4:0x8], byteorder='little')

    return size_compress, size_original


def is_efi_compressed(data: bytes | bytearray, strict: bool = True) -> bool:
    """ Check if data is EFI compressed, controlling EOF padding """

    size_comp, size_orig = efi_compress_sizes(data=data)

    check_diff: bool = size_comp < size_orig

    if strict:
        check_size: bool = size_comp + 0x8 == len(data)
    else:
        check_size = size_comp + 0x8 <= len(data)

    return check_diff and check_size


def efi_decompress(in_path: str, out_path: str, padding: int = 0, silent: bool = False,
                   comp_type: str = '--uefi') -> bool:
    """ EFI/Tiano Decompression via TianoCompress """

    try:
        subprocess.run(args=[tiano_path(), '-d', in_path, '-o', out_path, '-q', comp_type],
                       check=True, stdout=subprocess.DEVNULL)

        with open(file=in_path, mode='rb') as file:
            _, size_orig = efi_compress_sizes(data=file.read())

        if os.path.getsize(out_path) != size_orig:
            raise OSError('EFI decompressed file & header size mismatch!')
    except Exception as error:  # pylint: disable=broad-except
        if not silent:
            printer(message=f'Error: TianoCompress could not extract file {in_path}: {error}!', padding=padding)

        return False

    if not silent:
        printer(message='Successful EFI decompression via TianoCompress!', padding=padding)

    return True



#!/usr/bin/env python3 -B
# coding=utf-8

"""
Apple EFI ID
Apple EFI Image Identifier
Copyright (C) 2018-2024 Plato Mavropoulos
"""

import ctypes
import logging
import os
import struct
import subprocess
import zlib

from collections import defaultdict
from re import Match
from typing import Any, Final

from biosutilities.common.externals import uefiextract_path, uefifind_path
from biosutilities.common.paths import delete_dirs, delete_file, is_access, is_file, path_suffixes, runtime_root
from biosutilities.common.patterns import PAT_INTEL_IBIOSI, PAT_APPLE_ROM_VER
from biosutilities.common.structs import CHAR, ctypes_struct, UINT8
from biosutilities.common.system import printer
from biosutilities.common.templates import BIOSUtility
from biosutilities.common.texts import file_to_bytes


class IntelBiosId(ctypes.LittleEndianStructure):
    """
    Intel BIOS ID Structure

    https://github.com/tianocore/edk2-platforms/blob/master/Platform/Intel/BoardModulePkg/Include/Guid/BiosId.h
    """

    _pack_ = 1
    _fields_ = [
        ('Signature',       CHAR * 8),      # 0x00
        ('BoardID',         UINT8 * 16),    # 0x08
        ('Dot1',            UINT8 * 2),     # 0x18
        ('BoardExt',        UINT8 * 6),     # 0x1A
        ('Dot2',            UINT8 * 2),     # 0x20
        ('VersionMajor',    UINT8 * 8),     # 0x22
        ('Dot3',            UINT8 * 2),     # 0x2A
        ('BuildType',       UINT8 * 2),     # 0x2C
        ('VersionMinor',    UINT8 * 4),     # 0x2E
        ('Dot4',            UINT8 * 2),     # 0x32
        ('Year',            UINT8 * 4),     # 0x34
        ('Month',           UINT8 * 4),     # 0x38
        ('Day',             UINT8 * 4),     # 0x3C
        ('Hour',            UINT8 * 4),     # 0x40
        ('Minute',          UINT8 * 4),     # 0x44
        ('NullTerminator',  UINT8 * 2)      # 0x48
        # 0x4A
    ]

    @staticmethod
    def _decode(field: bytes) -> str:
        return struct.pack('B' * len(field), *field).decode(encoding='utf-16', errors='ignore').strip('\x00 ')

    def get_bios_id(self) -> dict[str, str]:
        """ Get Apple/Intel EFI BIOS ID """

        intel_sig: str = self.Signature.decode(encoding='utf-8')

        board_id: str = self._decode(field=self.BoardID)
        board_ext: str = self._decode(field=self.BoardExt)
        version_major: str = self._decode(field=self.VersionMajor)
        build_type: str = self._decode(field=self.BuildType)
        version_minor: str = self._decode(field=self.VersionMinor)
        build_year: str = self._decode(field=self.Year)
        build_month: str = self._decode(field=self.Month)
        build_day: str = self._decode(field=self.Day)
        build_hour: str = self._decode(field=self.Hour)
        build_minute: str = self._decode(field=self.Minute)

        efi_name_id: str = (f'{board_id}_{board_ext}_{version_major}_{build_type}{version_minor}'
                            f'_20{build_year}-{build_month}-{build_day}_{build_hour}-{build_minute}')

        return {
            'intel_sig': intel_sig,
            'board_id': board_id,
            'board_ext': board_ext,
            'version_major': version_major,
            'version_minor': version_minor,
            'build_type': build_type,
            'build_year': build_year,
            'build_month': build_month,
            'build_day': build_day,
            'build_hour': build_hour,
            'build_minute': build_minute,
            'efi_name_id': efi_name_id
        }

    def struct_print(self, padding: int = 0) -> None:
        """ Display structure information """

        ibiosi: dict[str, str] = self.get_bios_id()

        ibiosi_date: str = f'20{ibiosi["build_year"]}-{ibiosi["build_month"]}-{ibiosi["build_day"]}'
        ibiosi_time: str = f'{ibiosi["build_hour"]}:{ibiosi["build_minute"]}'

        printer(message=['Intel Signature:', ibiosi['intel_sig']], padding=padding, new_line=False)
        printer(message=['Board Identity: ', ibiosi['board_id']], padding=padding, new_line=False)
        printer(message=['Apple Identity: ', ibiosi['board_ext']], padding=padding, new_line=False)
        printer(message=['Major Version:  ', ibiosi['version_major']], padding=padding, new_line=False)
        printer(message=['Minor Version:  ', ibiosi['version_minor']], padding=padding, new_line=False)
        printer(message=['Build Type:     ', ibiosi['build_type']], padding=padding, new_line=False)
        printer(message=['Build Date:     ', ibiosi_date], padding=padding, new_line=False)
        printer(message=['Build Time:     ', ibiosi_time], padding=padding, new_line=False)


class AppleEfiIdentify(BIOSUtility):
    """ Apple EFI Image Identifier """

    TITLE: str = 'Apple EFI Image Identifier'

    ARGUMENTS: list[tuple[list[str], dict[str, str]]] = [
        (['-q', '--silent'], {'help': 'suppress structure display', 'action': 'store_true'})
    ]

    PAT_UEFIFIND: Final[str] = f'244942494F534924{"." * 32}2E00{"." * 12}2E00{"." * 16}2E00{"." * 12}2E00{"." * 40}00'

    def __init__(self, arguments: list[str] | None = None) -> None:
        super().__init__(arguments=arguments)

        self.efi_file_name: str = ''
        self.intel_bios_info: dict[str, str] = {}
        self.apple_rom_version: defaultdict[str, set] = defaultdict(set)

    def check_format(self, input_object: str | bytes | bytearray) -> bool:
        """ Check if input is Apple EFI image """

        if isinstance(input_object, str) and is_file(in_path=input_object) and is_access(in_path=input_object):
            if path_suffixes(in_path=input_object)[-1].lower() not in ('.fd', '.scap', '.im4p'):
                return False

            input_path: str = input_object
            input_buffer: bytes = file_to_bytes(in_object=input_path)
        elif isinstance(input_object, (bytes, bytearray)):
            input_path = os.path.join(runtime_root(), 'APPLE_EFI_ID_INPUT_BUFFER_CHECK.tmp')
            input_buffer = input_object

            with open(file=input_path, mode='wb') as check_out:
                check_out.write(input_buffer)
        else:
            return False

        try:
            if PAT_INTEL_IBIOSI.search(string=input_buffer):
                return True

            _ = subprocess.run([uefifind_path(), input_path, 'body', 'list', self.PAT_UEFIFIND],
                               check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

            return True
        except Exception as error:  # pylint: disable=broad-except
            logging.debug('Error: Could not check if input is Apple EFI image: %s', error)

            return False
        finally:
            if input_path != input_object:
                delete_file(in_path=input_path)

    def parse_format(self, input_object: str | bytes | bytearray, extract_path: str, padding: int = 0) -> bool:
        """ Parse & Identify (or Rename) Apple EFI image """

        input_buffer: bytes = file_to_bytes(in_object=input_object)

        if isinstance(input_object, str) and is_file(in_path=input_object):
            input_path: str = input_object
        else:
            input_path = os.path.join(runtime_root(), 'APPLE_EFI_ID_INPUT_BUFFER_PARSE.bin')

            with open(file=input_path, mode='wb') as parse_out:
                parse_out.write(input_buffer)

        bios_id_match: Match[bytes] | None = PAT_INTEL_IBIOSI.search(string=input_buffer)

        if bios_id_match:
            bios_id_res: str = f'0x{bios_id_match.start():X}'

            bios_id_hdr: Any = ctypes_struct(buffer=input_buffer, start_offset=bios_id_match.start(),
                                             class_object=IntelBiosId)
        else:
            # The $IBIOSI$ pattern is within EFI compressed modules so we need to use UEFIFind and UEFIExtract
            try:
                bios_id_res = subprocess.check_output([uefifind_path(), input_path, 'body', 'list',
                                                       self.PAT_UEFIFIND], text=True)[:36]

                # UEFIExtract must create its output folder itself
                delete_dirs(in_path=extract_path)

                _ = subprocess.run([uefiextract_path(), input_path, bios_id_res, '-o', extract_path, '-m', 'body'],
                                   check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

                with open(file=os.path.join(extract_path, 'body.bin'), mode='rb') as raw_body:
                    body_buffer: bytes = raw_body.read()

                # Detect decompressed $IBIOSI$ pattern
                bios_id_match = PAT_INTEL_IBIOSI.search(string=body_buffer)

                if not bios_id_match:
                    raise RuntimeError('Failed to detect decompressed $IBIOSI$ pattern!')

                bios_id_hdr = ctypes_struct(buffer=body_buffer, start_offset=bios_id_match.start(),
                                            class_object=IntelBiosId)

                delete_dirs(in_path=extract_path)  # Successful UEFIExtract extraction, remove its output folder
            except Exception as error:  # pylint: disable=broad-except
                printer(message=f'Error: Failed to parse compressed $IBIOSI$ pattern: {error}!', padding=padding)

                return False

        if not self.arguments.silent:
            printer(message=f'Detected Intel BIOS Info at {bios_id_res}\n', padding=padding)

            bios_id_hdr.struct_print(padding=padding + 4)

        self.intel_bios_info = bios_id_hdr.get_bios_id()

        self.efi_file_name = (f'{self.intel_bios_info["efi_name_id"]}_{zlib.adler32(input_buffer):08X}'
                              f'{path_suffixes(in_path=input_path)[-1]}')

        _ = self._apple_rom_version(input_buffer=input_buffer, padding=padding)

        if input_path != input_object:
            delete_file(in_path=input_path)

        return True

    def _apple_rom_version(self, input_buffer: bytes | bytearray, padding: int = 0) -> bool:
        rom_version_match: Match[bytes] | None = PAT_APPLE_ROM_VER.search(string=input_buffer)

        if rom_version_match:
            rom_version_match_off: int = rom_version_match.start()

            rom_version_header_len: int = input_buffer[rom_version_match_off:].find(b'\n')

            if rom_version_header_len != -1:
                rom_version_data_bgn: int = rom_version_match_off + rom_version_header_len

                rom_version_data_len: int = input_buffer[rom_version_data_bgn:].find(b'\x00')

                if rom_version_data_len != -1:
                    rom_version_data_end: int = rom_version_data_bgn + rom_version_data_len

                    rom_version_data: bytes = input_buffer[rom_version_data_bgn:rom_version_data_end]

                    rom_version_text: str = rom_version_data.decode('utf-8').strip('\n')

                    for rom_version_line in [line.strip() for line in rom_version_text.split('\n')]:
                        rom_version_parts: list[str] = rom_version_line.split(sep=':', maxsplit=1)

                        self.apple_rom_version[rom_version_parts[0].strip()].add(rom_version_parts[1].strip())

                    if not self.arguments.silent:
                        printer(message=f'Detected Apple ROM Version at 0x{rom_version_match_off:X}', padding=padding)

                        printer(message=rom_version_text, strip=True, padding=padding + 4)

                    return True

        return False


if __name__ == '__main__':
    AppleEfiIdentify().run_utility()



#!/usr/bin/env python3 -B
# coding=utf-8

"""
Apple EFI PKG
Apple EFI Package Extractor
Copyright (C) 2019-2024 Plato Mavropoulos
"""

import os

from biosutilities.common.compression import is_szip_supported, szip_decompress
from biosutilities.common.paths import (copy_file, delete_dirs, extract_folder, is_access, is_file, make_dirs,
                                        path_files, path_name, path_parent, path_suffixes, runtime_root)
from biosutilities.common.system import printer
from biosutilities.common.templates import BIOSUtility
from biosutilities.common.texts import file_to_bytes

from biosutilities.apple_efi_id import AppleEfiIdentify
from biosutilities.apple_efi_im4p import AppleEfiIm4pSplit
from biosutilities.apple_efi_pbzx import AppleEfiPbzxExtract


class AppleEfiPkgExtract(BIOSUtility):
    """ Apple EFI Package Extractor """

    TITLE: str = 'Apple EFI Package Extractor'

    def check_format(self, input_object: str | bytes | bytearray) -> bool:
        """ Check if input is Apple EFI PKG package """

        is_apple_efi_pkg: bool = False

        input_buffer: bytes = file_to_bytes(in_object=input_object)

        if isinstance(input_object, str) and is_file(in_path=input_object):
            input_path: str = input_object
        else:
            input_path = os.path.join(runtime_root(), 'APPLE_EFI_PKG_INPUT_BUFFER_CHECK.bin')

            with open(file=input_path, mode='wb') as input_path_object:
                input_path_object.write(input_buffer)

        for pkg_type in ('XAR', 'TAR', 'DMG'):
            if is_szip_supported(in_path=input_path, args=[f'-t{pkg_type}:s0']):
                is_apple_efi_pkg = True

                break

        if input_path != input_object:
            os.remove(path=input_path)

        return is_apple_efi_pkg

    def parse_format(self, input_object: str | bytes | bytearray, extract_path: str, padding: int = 0) -> bool:
        """ Parse & Extract Apple EFI PKG packages """

        if isinstance(input_object, str) and is_file(in_path=input_object):
            input_path: str = input_object
        else:
            input_path = os.path.join(runtime_root(), 'APPLE_EFI_PKG_INPUT_BUFFER_PARSE.bin')

            with open(file=input_path, mode='wb') as input_path_object:
                input_path_object.write(file_to_bytes(in_object=input_object))

        make_dirs(in_path=extract_path, delete=True)

        working_dir: str = os.path.join(extract_path, 'temp')

        make_dirs(in_path=working_dir)

        for pkg_type in ('XAR', 'TAR', 'DMG'):
            if is_szip_supported(in_path=input_path, padding=padding, args=[f'-t{pkg_type}']):
                if szip_decompress(in_path=input_path, out_path=working_dir, in_name=pkg_type, padding=padding,
                                   args=None if pkg_type == 'DMG' else [f'-t{pkg_type}']):
                    break
        else:
            return False

        if input_path != input_object:
            os.remove(path=input_path)

        for work_file in path_files(in_path=working_dir):
            if is_file(in_path=work_file) and is_access(in_path=work_file):
                self._pbzx_zip(input_path=work_file, extract_path=extract_path, padding=padding + 4)
                self._gzip_cpio(input_path=work_file, extract_path=extract_path, padding=padding + 4)
                self._dmg_zip(input_path=work_file, extract_path=extract_path, padding=padding + 4)
                self._xar_gzip(input_path=work_file, extract_path=extract_path, padding=padding + 4)

        delete_dirs(in_path=working_dir)

        return True

    def _xar_gzip(self, input_path: str, extract_path: str, padding: int = 0) -> None:
        """ XAR/TAR > GZIP """

        for pkg_type in ('XAR', 'TAR'):
            if is_szip_supported(in_path=input_path, padding=padding, args=[f'-t{pkg_type}']):
                pkg_path: str = extract_folder(in_path=input_path, suffix=f'_{pkg_type.lower()}_gzip')

                if szip_decompress(in_path=input_path, out_path=pkg_path, in_name=pkg_type,
                                   padding=padding, args=[f'-t{pkg_type}']):
                    for pkg_file in path_files(in_path=pkg_path):
                        if is_file(in_path=pkg_file) and is_access(in_path=pkg_file):
                            self._gzip_cpio(input_path=pkg_file, extract_path=extract_path, padding=padding + 4)

                break

    def _dmg_zip(self, input_path: str, extract_path: str, padding: int = 0) -> None:
        """ DMG > ZIP """

        if is_szip_supported(in_path=input_path, padding=padding, args=['-tDMG']):
            dmg_path: str = extract_folder(in_path=input_path, suffix='_dmg_zip')

            if szip_decompress(in_path=input_path, out_path=dmg_path, in_name='DMG', padding=padding, args=None):
                for dmg_file in path_files(in_path=dmg_path):
                    if is_file(in_path=dmg_file) and is_access(in_path=dmg_file):
                        if is_szip_supported(in_path=dmg_file, padding=padding + 4, args=['-tZIP']):
                            zip_path: str = extract_folder(in_path=dmg_file)

                            if szip_decompress(in_path=dmg_file, out_path=zip_path, in_name='ZIP',
                                               padding=padding + 4, args=['-tZIP']):
                                for zip_file in path_files(in_path=zip_path):
                                    self._im4p_id(input_path=zip_file, output_path=extract_path, padding=padding + 8)

    def _pbzx_zip(self, input_path: str, extract_path: str, padding: int = 0) -> None:
        """ PBZX > ZIP """

        pbzx_module: AppleEfiPbzxExtract = AppleEfiPbzxExtract()

        if pbzx_module.check_format(input_object=input_path):
            printer(message=f'Extracting PBZX via {pbzx_module.title}', padding=padding)

            pbzx_path: str = extract_folder(in_path=input_path, suffix='_pbzx_zip')

            if pbzx_module.parse_format(input_object=input_path, extract_path=pbzx_path, padding=padding + 4):
                printer(message=f'Successful PBZX extraction via {pbzx_module.title}!', padding=padding)

                for pbzx_file in path_files(in_path=pbzx_path):
                    if is_file(in_path=pbzx_file) and is_access(in_path=pbzx_file):
                        if is_szip_supported(in_path=pbzx_file, padding=padding + 4, args=['-tZIP']):
                            zip_path: str = extract_folder(in_path=pbzx_file)

                            if szip_decompress(in_path=pbzx_file, out_path=zip_path, in_name='ZIP',
                                               padding=padding + 4, args=['-tZIP']):
                                for zip_file in path_files(in_path=zip_path):
                                    self._im4p_id(input_path=zip_file, output_path=extract_path, padding=padding + 8)

    def _gzip_cpio(self, input_path: str, extract_path: str, padding: int = 0) -> None:
        """ GZIP > CPIO """

        if is_szip_supported(in_path=input_path, padding=padding, args=['-tGZIP']):
            gzip_path: str = extract_folder(in_path=input_path, suffix='_gzip_cpio')

            if szip_decompress(in_path=input_path, out_path=gzip_path, in_name='GZIP',
                               padding=padding, args=['-tGZIP']):
                for gzip_file in path_files(in_path=gzip_path):
                    if is_file(in_path=gzip_file) and is_access(in_path=gzip_file):
                        if is_szip_supported(in_path=gzip_file, padding=padding + 4, args=['-tCPIO']):
                            cpio_path: str = extract_folder(in_path=gzip_file)

                            if szip_decompress(in_path=gzip_file, out_path=cpio_path, in_name='CPIO',
                                               padding=padding + 4, args=['-tCPIO']):
                                for cpio_file in path_files(in_path=cpio_path):
                                    self._im4p_id(input_path=cpio_file, output_path=extract_path, padding=padding + 8)

    @staticmethod
    def _im4p_id(input_path: str, output_path: str, padding: int = 0) -> None:
        """ Split IM4P (if applicable), identify and rename EFI """

        if not (is_file(in_path=input_path) and is_access(in_path=input_path)):
            return None

        if path_suffixes(in_path=input_path)[-1].lower() not in ('.fd', '.scap', '.im4p'):
            return None

        if not AppleEfiIdentify().check_format(input_object=input_path):
            return None

        input_name: str = path_name(in_path=input_path)

        printer(message=input_name, padding=padding)

        working_dir: str = extract_folder(in_path=input_path)

        im4p_module: AppleEfiIm4pSplit = AppleEfiIm4pSplit()

        if im4p_module.check_format(input_object=input_path):
            printer(message=f'Splitting IM4P via {im4p_module.title}', padding=padding + 4)

            im4p_module.parse_format(input_object=input_path, extract_path=working_dir, padding=padding + 8)
        else:
            make_dirs(in_path=working_dir, delete=True)

            copy_file(in_path=input_path, out_path=working_dir, metadata=True)

        for efi_source in path_files(in_path=working_dir):
            if is_file(in_path=efi_source) and is_access(in_path=efi_source):
                efi_id_module: AppleEfiIdentify = AppleEfiIdentify()

                if efi_id_module.check_format(input_object=efi_source):
                    printer(message=f'Identifying EFI via {efi_id_module.title}', padding=padding + 4)

                    if efi_id_module.parse_format(input_object=efi_source, extract_path=extract_folder(
                            in_path=efi_source), padding=padding + 8):
                        efi_dest: str = os.path.join(path_parent(in_path=efi_source), efi_id_module.efi_file_name)

                        os.rename(src=efi_source, dst=efi_dest)

        for efi_final in path_files(in_path=working_dir):
            if is_file(in_path=efi_final) and is_access(in_path=efi_final):
                copy_file(in_path=efi_final, out_path=output_path, metadata=True)

        delete_dirs(in_path=working_dir)

        return None


if __name__ == '__main__':
    AppleEfiPkgExtract().run_utility()



#!/usr/bin/env python3 -B
# coding=utf-8

"""
Fujitsu UPC Extract
Fujitsu UPC BIOS Extractor
Copyright (C) 2021-2024 Plato Mavropoulos
"""

import os

from biosutilities.common.compression import efi_decompress, is_efi_compressed
from biosutilities.common.paths import make_dirs, is_file, path_name, path_suffixes
from biosutilities.common.templates import BIOSUtility
from biosutilities.common.texts import file_to_bytes


class FujitsuUpcExtract(BIOSUtility):
    """ Fujitsu UPC BIOS Extractor """

    TITLE: str = 'Fujitsu UPC BIOS Extractor'

    def check_format(self, input_object: str | bytes | bytearray) -> bool:
        """ Check if input is Fujitsu UPC image """
        is_upc: bool = False

        if isinstance(input_object, str) and is_file(in_path=input_object):
            is_upc = path_suffixes(input_object)[-1].upper() == '.UPC'
        elif isinstance(input_object, (bytes, bytearray)):
            is_upc = True

        if is_upc:
            is_upc = is_efi_compressed(data=file_to_bytes(in_object=input_object))

        return is_upc

    def parse_format(self, input_object: str | bytes | bytearray, extract_path: str, padding: int = 0) -> bool:
        """ Parse & Extract Fujitsu UPC image """

        make_dirs(in_path=extract_path, delete=True)

        if isinstance(input_object, str) and is_file(in_path=input_object):
            input_name: str = path_name(in_path=input_object)

            input_path: str = input_object

            if input_name.upper().endswith('.UPC'):
                input_name = input_name[:-4]
        else:
            input_name = 'Fujitsu_UPC_Image'

            input_path = os.path.join(extract_path, f'{input_name}.UPC')

            with open(file=input_path, mode='wb') as input_path_object:
                input_path_object.write(file_to_bytes(in_object=input_object))

        output_path: str = os.path.join(extract_path, f'{input_name}.bin')

        efi_status: bool = efi_decompress(in_path=input_path, out_path=output_path, padding=padding)

        if input_path != input_object:
            os.remove(path=input_path)

        return efi_status


if __name__ == '__main__':
    FujitsuUpcExtract().run_utility()



#!/usr/bin/env python3 -B
# coding=utf-8

"""
Copyright (C) 2022-2024 Plato Mavropoulos
"""

import os
import sys

from argparse import ArgumentParser, Namespace
from typing import Final

from biosutilities import __version__
from biosutilities.common.paths import (delete_dirs, extract_folder, is_access, is_dir, is_file, is_empty_dir,
                                        path_files, path_name, path_parent, real_path, runtime_root)
from biosutilities.common.system import system_platform, python_version, printer
from biosutilities.common.texts import remove_quotes, to_boxed, to_ordinal


class BIOSUtility:
    """ Base utility class for BIOSUtilities """

    TITLE: str = 'BIOS Utility'

    ARGUMENTS: list[tuple[list[str], dict[str, str]]] = []

    MAX_FAT32_ITEMS: Final[int] = 65535

    MIN_PYTHON_VER: Final[tuple[int, int]] = (3, 10)

    def __init__(self, arguments: list[str] | None = None) -> None:
        self.title: str = f'{self.TITLE.strip()} v{__version__}'

        argparser: ArgumentParser = ArgumentParser(allow_abbrev=False)

        argparser.add_argument('paths', nargs='*')
        argparser.add_argument('-e', '--auto-exit', help='skip user action prompts', action='store_true')
        argparser.add_argument('-o', '--output-dir', help='output extraction directory')

        for argument in self.ARGUMENTS:
            argparser.add_argument(*argument[0], **argument[1])  # type: ignore

        sys_argv: list[str] = arguments if isinstance(arguments, list) and arguments else sys.argv[1:]

        self.arguments: Namespace = argparser.parse_known_args(sys_argv)[0]

        self._input_files: list[str] = []

        self._output_path: str = ''

    def run_utility(self, padding: int = 0) -> bool:
        """ Run utility after checking for supported format """

        self._check_sys_py()

        self._check_sys_os()

        self.show_version(padding=padding)

        self._setup_input_files(padding=padding)

        self._setup_output_dir(padding=padding)

        exit_code: int = len(self._input_files)

        for input_file in self._input_files:
            input_name: str = path_name(in_path=input_file, limit=True)

            printer(message=input_name, padding=padding + 4)

            if not self.check_format(input_object=input_file):
                printer(message='Error: This is not a supported format!', padding=padding + 8)

                continue

            extract_path: str = os.path.join(self._output_path, extract_folder(in_path=input_name))

            if is_dir(in_path=extract_path):
                for suffix in range(2, self.MAX_FAT32_ITEMS):
                    renamed_path: str = f'{os.path.normpath(path=extract_path)}_{to_ordinal(in_number=suffix)}'

                    if not is_dir(in_path=renamed_path):
                        extract_path = renamed_path

                        break

            if self.parse_format(input_object=input_file, extract_path=extract_path, padding=padding + 8):
                exit_code -= 1

            if is_empty_dir(in_path=extract_path):
                delete_dirs(in_path=extract_path)

        printer(message='Done!\n' if not self.arguments.auto_exit else None, pause=not self.arguments.auto_exit)

        return exit_code == 0

    def show_version(self, is_boxed: bool = True, padding: int = 0) -> None:
        """ Show title and version of utility """

        printer(message=to_boxed(in_text=self.title) if is_boxed else self.title, new_line=False, padding=padding)

    def parse_format(self, input_object: str | bytes | bytearray, extract_path: str, padding: int = 0) -> bool:
        """ Process input object as a specific supported format """

        raise NotImplementedError(f'Method "parse_format" not implemented at {__name__}')

    def check_format(self, input_object: str | bytes | bytearray) -> bool:
        """ Check if input object is of specific supported format """

        raise NotImplementedError(f'Method "check_format" not implemented at {__name__}')

    def _setup_input_files(self, padding: int = 0) -> None:
        self._input_files = []

        input_paths: list[str] = self.arguments.paths

        if not input_paths:
            input_paths = [remove_quotes(in_text=input(f'\n{" " * padding}Enter input file or directory path: '))]

        for input_path in [input_path for input_path in input_paths if input_path]:
            input_path_real: str = real_path(in_path=input_path)

            if is_dir(in_path=input_path_real):
                for input_file in path_files(in_path=input_path_real):
                    if is_file(in_path=input_file) and is_access(in_path=input_file):
                        self._input_files.append(input_file)
            elif is_file(in_path=input_path_real) and is_access(in_path=input_path_real):
                self._input_files.append(input_path_real)

    def _setup_output_dir(self, padding: int = 0) -> None:
        self._output_path = ''

        output_path: str = self.arguments.output_dir

        if not output_path:
            output_path = remove_quotes(in_text=input(f'\n{" " * padding}Enter output directory path: '))

            if not output_path and self._input_files:
                output_path = str(path_parent(in_path=self._input_files[0]))

        if output_path and is_dir(in_path=output_path) and is_access(in_path=output_path):
            self._output_path = output_path
        else:
            self._output_path = runtime_root()

    def _check_sys_py(self) -> None:
        """ Check Python Version """

        sys_py: tuple = python_version()

        if sys_py < self.MIN_PYTHON_VER:
            min_py_str: str = '.'.join(map(str, self.MIN_PYTHON_VER))
            sys_py_str: str = '.'.join(map(str, sys_py[:2]))

            raise RuntimeError(f'Python >= {min_py_str} required, not {sys_py_str}')

    @staticmethod
    def _check_sys_os() -> None:
        """ Check OS Platform """

        os_tag, is_win, is_lnx = system_platform()

        if not (is_win or is_lnx):
            raise OSError(f'Unsupported operating system: {os_tag}')
